// $Id: hash.mq4 84 2014-02-07 02:10:06Z ydrol $
//#property strict

/*
   This is losely ported from a C version I have which was in turn modified from hashtable.c by Christopher Clark.
 Copyright (C) 2014, Andrew Lord (NICKNAME=lordy) <forex@NICKNAME.org.uk> 
 Copyright (C) 2002, 2004 Christopher Clark <firstname.lastname@cl.cam.ac.uk> 

 ---------------

 This creates an array of linked lists (HashEntry).
 There is one list for each hash value.
 To store an object against a string(key) - get the string hashcode, then insert pair (key,val) into the linked list for that hashcode.
 To fetch an object against a string(key) - get the string hashcode, get linked-list at that hashcode index, then search for the key and return the val.

the Objects to store must be a sub class of HashValue. - This might be a problem in MQL if your object is already a subclass
as there is no multiple inheritance. Happy for suggestions to improve this...

There are some convenience classes to hold atomic types as values HashString,HashDouble,HashInt

EXAMPLE:


class myClass: public HashValue {
   public: int v;
   myClass(int a) { v = a;}
};

// Create the objects as needed

      myClass *a = new myClass(1);
      myClass *b = new myClass(2);
      myClass *c = new myClass(3);

// Then to insert into hash etc.

      Hash* h = new Hash(256,true); 
      // 'true' means when the hash will adopt the values and delete them when they are removed from the hash or when the hash is deleted.

      h.hPut("a",a);
      h.hPut("b",b);
      h.hPut("c",c);

      myClass *d = h.hGet("b");

      etc.

// Iterate over hash
    HashLoop *l
    for (l = new HashLoop(h) ; l.hasNext() ; l.next()  ) {
        string key = l.key();
        MyClass *c = l.val();
    }
    delete l;

    // Delete from hash - This will also delete 'a' because we set the 'adopt' flag on the hash.
    h.hDel("a");

    //Delete the hash - this will also delete 'b' and 'c' because of the adopt flag.
    delete h;
 
      
*/

// Any value stored in a Hash must be a subclass of HashValue
class HashValue {
};

// Linked list of values - there will be one list for each hash value
class HashEntry {
    public:
        string _key;
        HashValue * _val;
        HashEntry *_next;

        HashEntry() {
            _key=NULL;
            _val=NULL;
            _next=NULL;
        }

        HashEntry(string key,HashValue* val) {
            _key=key;
            _val=val;
            _next=NULL;
        }

        ~HashEntry() {
        }
};

// Convenience classes for storing atomic types as hash values.
class HashString : public HashValue { public: string val; HashString(string v) { val=v;} };
class HashDouble : public HashValue { public: double val;  HashDouble(double v) { val=v;}};
class HashInt : public HashValue { public: int val;  HashInt(int v) { val=v;}};
class HashDatetime : public HashValue { public: datetime val;  HashDatetime(datetime v) { val=v;}};
/*
   Main hash class. It is also a subclass of HashValue so Hashes of Hashes are possible
*/
class Hash : public HashValue {

private:
    uint _hashSlots;     // Number of slots in the hashtable. 
                    // this should be approx number of elements to store. Depending on hash algorithm
                    // it may optimally be a prime or a power of two etc. but probably not important
                    // unless hashes are massive. Best optimisation at present is to move the hashcode
                    // function to a DLL??

    int _hashEntryCount; // number of things in the hash
    HashEntry* _table[];

    bool _adoptValues;

    int _errCode;
    string _errText;

    void init(int size,bool adoptValues) {
        _hashSlots = 0;
        _hashEntryCount = 0;
        clearErr();
        setAdoptValues(adoptValues);
        int n = ArrayResize(_table,size);
        if (n >= 1) {
            _hashSlots = size;
            for(int i = 0 ; i < n ; i++ ) _table[i] = NULL;
        } else {
            setErr(-1,"unable to init");
        }
    }

public:

    int getErrCode() {
        return _errCode;
    }
    string getErrText() {
        return _errText;
    }
    void clearErr() {
        _errCode=0;
        _errText="";
    }
    void setErr(int e,string m) {
        _errCode = e;
        _errText = m;
        //error((string)e,m);
    }

    uint hash(string s) {
        return rawhash(s) % _hashSlots;
    }

    // this is the bottleneck function. Shame mql hash no default hash method for objects.
    // After reviewing quite a few hash functions I settled on the one below.
    // http://www.cse.yorku.ca/~oz/hash.html
    static uint rawhash(string s) {
        uchar c[];
        uint h = 0;

        if (s != NULL) {
            h = 5381;
            int n = StringToCharArray(s,c);
            for(int i = 0 ; i < n ; i++ ) {
                h = ((h << 5 ) + h ) + c[i];
            }
        }
        return h;
    }

    Hash() {
        init(256,true);
    }

    Hash(int size,bool adoptValues) {
        init(size,adoptValues);
    }

    ~Hash() {

        // Free entries.
        for(uint i = 0 ; i< _hashSlots ; i++) {
            HashEntry *nextEntry = NULL;
            for(HashEntry *entry = _table[i] ; entry!= NULL ; entry = nextEntry ) 
            {
                nextEntry = entry._next;

                if (_adoptValues && entry._val != NULL) {
                    delete entry._val;
                }
                delete entry;
            }
            _table[i] = NULL;
        }
    }

    void setAdoptValues(bool v) {
        _adoptValues = v;
    }

    bool getAdoptValues() {
        return _adoptValues;
    }

    private:
    uint _foundIndex;     // After find() is called is set to hashindex for name whether found or not.
    HashEntry* _foundEntry;// After find() is called  is set to the HashEntry that contains the key.
    HashEntry* _foundPrev;// After find() is called  is set to the HashEntry before the entry
                            // (could use double linked list but requires more memory).

    // Look for the required entry for key 'name' true if found.
    bool find(string name) {
    
         //Alert("finding");
        bool found = false;

        // Get the index using the hashcode of the string
        _foundIndex = hash(name);
        

        if (_foundIndex>_hashSlots ) {

            setErr(1,"hGet: bad hashIndex="+(string)_foundIndex+" size "+(string)_hashSlots);

        } else {

            // Search the linked list determined by the index.
            
            for(HashEntry *e = _table[_foundIndex] ; e != NULL ; e = e._next )  {
                if (e._key == name) {
                    _foundEntry = e;
                    found=true;
                    break;
                }
                // Track the item before the target item in case deleting from single linked list.
                _foundPrev = e;
            }
        }

        return found;
    }

    public:

    // This is used by the HashLoop class
    HashEntry*getEntry(int i) {
        return _table[i];
    }

    // Return the number of slots (not number of elements)
    uint getSlots() {
        return _hashSlots;
    }
    int getCount() {
        return _hashEntryCount;
    }

    bool hContainsKey(string name) {
        return find(name);
    }

    //Fetch a value using string key
    //NULL=not found (if not storing nulls)
    //
    //If not storing nulls use
    // obj = hash.hGet(x); if (obj != NULL) OK
    //
    //If storing nulls use
    // obj = hash.hGet(x); if (obj != NULL || hash.errCode() == 0 ) OK
    HashValue* hGet(string name) {
        HashValue *obj = NULL;
        clearErr();
        bool found=false;

        if (find(name)) {
            obj = _foundEntry._val;
        } else {
            //If Hash contains nulls then also check the errorCode=0 when retrieving
            if (!found) {
                setErr(1,"not found");
            }
        }
        return obj;
    }

    //Convenience method for getting string values from hash of HashString
    string hGetString(string name) {
        string ret = NULL;
        HashString *v = hGet(name);
        if (v != NULL) {
            ret = v.val;
        }
        return ret;
    }
    //Convenience method for getting string values from hash of HashDouble
    double hGetDouble(string name) {
        double ret = NULL;
        HashDouble *v = hGet(name);
        if (v != NULL) {
            ret = v.val;
        }
        return ret;
    }
    //Convenience method for getting string values from hash of HashInt
    int hGetInt(string name) {
        int ret = NULL;
        HashInt *v = hGet(name);
        if (v != NULL) {
            ret = v.val;
        }
        return ret;
    }
    //Convenience method for getting string values from hash of HashDouble
    datetime hGetDatetime(string name) {
        datetime ret = NULL;
        HashDatetime *v = hGet(name);
        if (v != NULL) {
            ret = v.val;
        }
        return ret;
    }

    // This will overwrite existing
    void hPut(string name,HashValue *obj) {
    
         //Alert("clearing");
        clearErr();

         
        if (find(name)) {
            // Replace entry contents
            if (_adoptValues && _foundEntry._val != NULL) {
                delete _foundEntry._val;
            }
            _foundEntry._val = obj;

        } else {
            // Insert new entry at head of list
            HashEntry* e = new HashEntry(name,obj);
            HashEntry* first = _table[_foundIndex];
            e._next = first;
            _table[_foundIndex] = e;
            _hashEntryCount++;
        }
    }
    // Store a string as hash value
    void hPutString(string name,string s) {
        HashString *v = new HashString(s);
        hPut(name,v);
    }
    // Store a double as hash value 
    void hPutDouble(string name,double d) {
        HashDouble *v = new HashDouble(d);
        hPut(name,v);
    }
    // Store an int as hash value
    void hPutInt(string name,int i) {
        HashInt *v = new HashInt(i);
        hPut(name,v);
    }

    // Store a datetime as hash value
    void hPutDatetime(string name,datetime i) {
        HashDatetime *v = new HashDatetime(i);
        hPut(name,v);
    }

    // Delete and entry from the list.
    bool hDel(string name) {

        bool found = false;
        clearErr();

        if (find(name)) {
            HashEntry *next = _foundEntry._next;
            if (_foundPrev != NULL) {
                //Remove entry from the middle of the list.
                _foundPrev._next = next;
            } else {
                // remove from head of list
                _table[_foundIndex] = next;
            }

            if (_adoptValues && _foundEntry._val != NULL) delete _foundEntry._val;
            delete _foundEntry;
            _hashEntryCount--;
            found=true;

        }
        return found;
    }
};

/*=============================================
  Class to iterate over a Hash using ...

    HashLoop *l
    for (l = new HashLoop(h) ; l.hasNext() ; l.next()  ) {
        string key = l.key();
        MyClass *c = l.val();
    }
    delete l;

    Subclass of Hash to allow private member access

================================================*/
    class HashLoop : public Hash {
    private:
        uint _index;
        HashEntry *_currentEntry;
        Hash *_hash;

    public:
        HashLoop(Hash *h) {
            setHash(h);
        }
        ~HashLoop() {};

        void reset() {
            _index=0;
            _currentEntry = _hash.getEntry(_index);
            if (_currentEntry == NULL) {
                next();
            }
        }

        void setHash(Hash *h) {
            _hash = h;
            reset();
        }

        bool hasNext() {
            bool ret = ( _currentEntry != NULL);
            //config("hasNext=",ret);
            return ret;
        }

        void next() {

            //config("next : index = ",_index);

            // Advance
            if (_currentEntry != NULL) {
                _currentEntry = _currentEntry._next;
            }

            // Keep advancing if _currentEntry is null
            while (_currentEntry==NULL) {
                _index++;
                if (_index >= _hash.getSlots() ) return ;
                _currentEntry = _hash.getEntry(_index);
            }
        }

        // Return the key
        string key() {
            if (_currentEntry != NULL) {
                return _currentEntry._key;
            } else {
                return NULL;
            }
        }

        // Return the value
        HashValue *val() {
            if (_currentEntry != NULL) {
                return _currentEntry._val;
            } else {
                return NULL;
            }
        }

        // Convenience functions for retriving atomic types via loop
        int valInt() {
            return ((HashInt *)val()).val;
        }

        string valString() {
            return ((HashString *)val()).val;
        }

        double valDouble() {
            return ((HashDouble *)val()).val;
        }

        datetime valDatetime() {
            return ((HashDatetime *)val()).val;
        }
};


// Unit test - replace error() with Print(), Alert() or your own error logging.

void testHash() {

#define HASH_TEST_SIZE 100

    // Force hash collisions using a small hash size.
    Hash *h = new Hash(5,true);

    int count = 0;
    HashLoop *l ;

    for(int i = 0 ; i < HASH_TEST_SIZE ; i++ ) {
        h.hPutInt((string)i,i);
    }
    // Read and verify
    for(int i = 0 ; i < HASH_TEST_SIZE ; i++ ) {
        int j = h.hGetInt((string)i);
        if (i != j) error("Hash fail1",(string)i,(string)j);
    }
    count = 0;

    int a[HASH_TEST_SIZE];

    ArrayInitialize(a,0);
    // Test the loop
    for( l = new HashLoop(h) ; l.hasNext() ; l.next()) {
        count++;

        //Check loop key values match
        int i = (int)StringToInteger(l.key());
        int j = l.valInt();
        if (i != j) error("Hash fail2",(string)i,(string)j);

        // Check occurences
        a[i]++;
        if (a[i] != 1) error("Hash fail3",(string)i,(string)j);
    }
    delete l;

    // Check total
    if (count != HASH_TEST_SIZE) error("Hash fail4 ",(string)count);

    delete h;
}
